/** @file function.h
 *
 *  Interface to class of symbolic functions. */

/*
 *  GiNaC Copyright (C) 1999-2006 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __GINAC_FUNCTION_H__
#define __GINAC_FUNCTION_H__

#include "exprseq.h"

#include <iostream>

namespace GiNaC {

class function : public exprseq
{
	GINAC_DECLARE_REGISTERED_CLASS_NO_CTORS(function, exprseq)
public:
	virtual function* duplicate() const;
	virtual void accept(GiNaC::visitor& v) const;
	function(const exvector& v) : exprseq(v) { tinfo_key = &function::tinfo_static; }
	function(std::auto_ptr<exvector> vp) : exprseq(vp) { tinfo_key = &function::tinfo_static; }
	function(tinfo_t ti, const exvector& vp) : exprseq(vp) { tinfo_key = ti; }
	function(tinfo_t ti, std::auto_ptr<exvector> vp) : exprseq(vp) { tinfo_key = ti; }
	ex thiscontainer(const exvector& v) const { return function(v); }
	ex thiscontainer(std::auto_ptr<exvector> vp) const { return function(vp); }
protected:
	virtual int compare_same_type(const basic& other) const;
	virtual bool is_equal_same_type(const basic& other) const;
	virtual bool match_same_type(const basic & other) const;
	virtual unsigned return_type() const;
	virtual tinfo_t return_type_tinfo() const;
public:
	virtual ex conjugate() const;
	virtual ex derivative(const symbol& s) const;
	virtual ex pderivative(unsigned diff_param) const;
	virtual ex expand(unsigned options) const;
	virtual ex power_law(const ex& exp) const;
public:
	virtual bool info(unsigned inf) const;
	virtual void print(const print_context& c, unsigned level = 0) const;
public:
	function() { }
	function(tinfo_t ti) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1) : inherited(x1) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2) : inherited(x1, x2) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3) : inherited(x1, x2, x3) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4) : inherited(x1, x2, x3, x4) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5) : inherited(x1, x2, x3, x4, x5) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5, const ex& x6) : inherited(x1, x2, x3, x4, x5, x6) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5, const ex& x6, const ex& x7) : inherited(x1, x2, x3, x4, x5, x6, x7) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5, const ex& x6, const ex& x7, const ex& x8) : inherited(x1, x2, x3, x4, x5, x6, x7, x8) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5, const ex& x6, const ex& x7, const ex& x8, const ex& x9) : inherited(x1, x2, x3, x4, x5, x6, x7, x8, x9) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5, const ex& x6, const ex& x7, const ex& x8, const ex& x9, const ex& x10) : inherited(x1, x2, x3, x4, x5, x6, x7, x8, x9, x10) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5, const ex& x6, const ex& x7, const ex& x8, const ex& x9, const ex& x10, const ex& x11) : inherited(x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11) { tinfo_key = ti; }
	function(tinfo_t ti, const ex& x1, const ex& x2, const ex& x3, const ex& x4, const ex& x5, const ex& x6, const ex& x7, const ex& x8, const ex& x9, const ex& x10, const ex& x11, const ex& x12) : inherited(x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12) { tinfo_key = ti; }
};

/** Macro for inclusion in the declaration of a GiNaC function. */
#define GINAC_DECLARE_FUNCTION_NONSTD(classname) \
public: \
	typedef function inherited; \
    static const tinfo_static_t tinfo_static; \
private: \
	static GiNaC::registered_class_info reg_info; \
public: \
	static GiNaC::registered_class_info& get_class_info_static() { return reg_info; } \
	virtual const GiNaC::registered_class_info& get_class_info() const { return classname::get_class_info_static(); } \
	virtual GiNaC::registered_class_info& get_class_info() { return classname::get_class_info_static(); } \
	virtual const char* class_name() const { return classname::get_class_info_static().options.get_name(); } \
	\
	classname(const GiNaC::archive_node& n, GiNaC::lst& sym_lst) : inherited(n, sym_lst) { } \
	virtual void archive(GiNaC::archive_node& n) const { inherited::archive(n); } \
	static GiNaC::ex unarchive(const GiNaC::archive_node& n, GiNaC::lst& sym_lst) { return (new classname(n, sym_lst))->setflag(status_flags::dynallocated); } \
	\
	static GiNaC::ex factory(const exvector& args) { return (new classname(args))->setflag(status_flags::dynallocated); } \
	\
	class visitor { \
	public: \
		virtual void visit(const classname&) = 0; \
		virtual ~visitor() {}; \
	}; \
	\
	virtual void accept(GiNaC::visitor& v) const \
	{ \
		if (visitor* p = dynamic_cast<visitor*>(&v)) \
			p->visit(*this); \
		else \
			inherited::accept(v); \
	} \
	\
	virtual classname* duplicate() const { return new classname(*this); } \
private:

#define GINAC_DECLARE_FUNCTION(classname) \
GINAC_DECLARE_FUNCTION_NONSTD(classname) \
public: \
	classname(const exvector& v) : inherited(v) { tinfo_key = &classname::tinfo_static; } \
	classname(std::auto_ptr<exvector> vp) : inherited(vp) { tinfo_key = &classname::tinfo_static; } \
	ex thiscontainer(const exvector& v) const { return classname(v); } \
	ex thiscontainer(std::auto_ptr<exvector> vp) const { return classname(vp); } \
private:

#define GINAC_FUNCTION_CTOR_1P(classname) \
	public: \
		classname(const ex& x1) : inherited(&classname::tinfo_static, x1) { }

#define GINAC_FUNCTION_CTOR_2P(classname) \
	public: \
		classname(const ex& x1, const ex& x2) : inherited(&classname::tinfo_static, x1, x2) { } \

#define GINAC_FUNCTION_CTOR_3P(classname) \
	public: \
		classname(const ex& x1, const ex& x2, const ex& x3) : inherited(&classname::tinfo_static, x1, x2, x3) { } \

/** Macro for inclusion in the declaration of a GiNaC function with one parameter. */
#define GINAC_DECLARE_FUNCTION_1P(classname) \
	GINAC_DECLARE_FUNCTION(classname) \
	GINAC_FUNCTION_CTOR_1P(classname) \
	private:

/** Macro for inclusion in the declaration of a GiNaC function with two parameters. */
#define GINAC_DECLARE_FUNCTION_2P(classname) \
	GINAC_DECLARE_FUNCTION(classname) \
	GINAC_FUNCTION_CTOR_2P(classname) \
	private:

/** Macro for inclusion in the declaration of a GiNaC function with two parameters. */
#define GINAC_DECLARE_FUNCTION_3P(classname) \
	GINAC_DECLARE_FUNCTION(classname) \
	GINAC_FUNCTION_CTOR_3P(classname) \
	private:

#define GINAC_FUNCTION_eval public: virtual ex eval(int level = 0) const;
#define GINAC_FUNCTION_evalf public: virtual ex evalf(int level = 0) const;
#define GINAC_FUNCTION_conjugate public: virtual ex conjugate() const;
#define GINAC_FUNCTION_power_law public: virtual ex power_law(const ex& exp) const;
#define GINAC_FUNCTION_series public: virtual ex series(const relational& r, int order, unsigned options = 0) const;
#define GINAC_FUNCTION_print_dflt protected: void do_print_dflt(const print_context& c, unsigned level) const;
#define GINAC_FUNCTION_print_csrc_float protected: void do_print_csrc_float(const print_context& c, unsigned level) const;
#define GINAC_FUNCTION_print_latex protected: void do_print_latex(const print_context& c, unsigned level) const;

#define GINAC_FUNCTION_1P(functionname, class_methods) \
	class functionname##_function : public function \
	{ \
	GINAC_DECLARE_FUNCTION_1P(functionname##_function) \
	class_methods \
	}; \
	template<typename T1> inline functionname##_function functionname(const T1& x1) { return functionname##_function(x1); }

#define GINAC_FUNCTION_2P(functionname, class_methods) \
	class functionname##_function : public function \
	{ \
	GINAC_DECLARE_FUNCTION_2P(functionname##_function) \
	class_methods \
	}; \
	template<typename T1, typename T2> inline functionname##_function functionname(const T1& x1, const T2& x2) { return functionname##_function(x1, x2); }

/** Macro for inclusion in the implementation of a GiNaC function. */
#define GINAC_IMPLEMENT_FUNCTION(classname) \
	GiNaC::registered_class_info classname::reg_info = GiNaC::registered_class_info(GiNaC::registered_class_options(#classname, "function", &classname::tinfo_static, &classname::unarchive).func_factory(&classname::factory)); \
	const tinfo_static_t classname::tinfo_static = {};
			
/** Macro for inclusion in the implementation of a GiNaC function.
 *  Additional options can be specified. */
#define GINAC_IMPLEMENT_FUNCTION_OPT(classname, options) \
	GiNaC::registered_class_info classname::reg_info = GiNaC::registered_class_info(GiNaC::registered_class_options(#classname, "function", &classname::tinfo_static, &classname::unarchive).func_factory(&classname::factory).options); \
	const tinfo_static_t classname::tinfo_static = {};

/** Exception class thrown by classes which provide their own series expansion
 *  to signal that ordinary Taylor expansion is safe. */
class do_taylor {};

} // namespace GiNaC

#endif // ifndef __GINAC_FUNCTION_H__
